#ifndef COMMON_H
#define COMMON_H
#include <cfloat> //DBL_MIN
#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include <deque>
#include <eigen3/Eigen/Dense>
#include <boost/math/special_functions/gamma.hpp>
using Eigen::MatrixXd;
using Eigen::VectorXd;
using Eigen::FullPivLU;
namespace common
{
    //subject_data_type is for phenotype
    enum subject_data_type
    {
        FULL_DATA,
        MISSING_DATA
    };


    //measure_type is for biomarkers
    enum measure_type
    {
        MEASURE_TYPE_KNOWN,
        MEASURE_TYPE_BELOW_DETECT_LIMIT,
        MEASURE_TYPE_ABOVE_DETECT_LIMIT,
        MEASURE_TYPE_MISSING_DATA
    };

    struct measurement
    {
        measure_type type;
        double measured_value;
        double upper_bound = DBL_MAX;
        double lower_bound = DBL_MIN;
    };

    //read_geno()
    //return value : error code as int
    //return NO_ERROR if no error
    //pre: filename exists and is in the genotype
    //file format described in documentation
    //pheno_subject_data_vec.size() is
    //initialized by being set in read_geno_and_biomarker()
    //pheno_subject_data_vec is used to remove any
    //subjects with missing phenotype from genotype 
    //information
    //the min_call_rate parameter is used to 
    //define the minimum call rate needed to include a SNP.
    //post: snp_vxd_vec contains the genotype data
    //for each SNP with call rate of at least min_call_rate
    //snp_name_vector contains names for the SNPs. 
    int read_geno( const std::string & filename, 
                    std::vector< VectorXd > & snp_vxd_vec,
                    std::vector<std::string> & snp_name_vector,
                    const std::vector<subject_data_type> & 
                                    pheno_subject_data_vec,
                    double min_call_rate );

    //read_configfile()
    //return value is error code as an int.  
    //If no error return NO_ERROR
    //pre: config_filename exists and 
    //is a configuration file in the format described in
    //the documentation
    //post: the non-constant reference
    //parameters are set using the values in
    //the config file config_filename
    int read_configfile( const std::string & config_filename,
                               std::string & geno_filename,
                               std::string & pheno_list_filename,
                               bool & pheno_covar_snp,
                               size_t & out_freq,
                               double & min_call_rate,
                               size_t & max_iter,
                               double & converged_ratio,
                               std::string & out_suffix,
                               bool & write_covariance,
                               bool & write_full_estimate );

    //read_pheno_and_biomarker()
    //return value : error code as int.  NO_ERROR 
    //if no errors.
    //pre: filename exists and has data in format
    //described in documentation.
    //post: Y_pheno_vec.rows() will equal
    //number of subjects with non-missing phenotypes.
    //covar_mxd.rows() == number of covariates + 1 (for constant)
    //covar_mxd.cols() == number of subjects with non-missing pheno
    //biomarker_measurements.size() == number of subjects with
    //non-missing phenotype.
    //covar_name_vec will contain the names of all covariates
    //(but constants will not be included)
    //subject_data_vec.size() will equal the number of subject lines
    //in the input filename, including missing phenotype subjects.  
    //subject_data_vec is intended to allow for the removal of
    //subjects with missing phenotype from the genotype data
    //read in later.
    int read_pheno_and_biomarker( const std::string & filename, 
                    std::vector<measurement> & biomarker_measurements,
                    VectorXd & Y_pheno_vec,
                    MatrixXd & covar_mxd,
                    std::vector< subject_data_type > & subject_data_vec,
                    std::vector< std::string > & covar_name_vec,
                    bool & isPhenoBinary );

    //calc_variance()
    //returns variance of the values in double vector
    //if double_vector is empty a message will be printed to 
    //output and 0.0 will be returned
    double calc_variance( const std::vector< double > & double_vector ); 

    //return true iff old_vec and new_vec are of same length and
    //each element of old_vec "agrees" with corresponding element of new_vec
    //to within convergence_ratio (simplified - see code for details)
    bool isVectorConverged( const VectorXd & old_vec, const VectorXd & new_vec,
                            double convergence_ratio );

    //make_mxd_with_constant_from_covar_mxd()
    //return value is error_code as int.  NO_ERROR if no error.
    //pre: covar_mxd : covariate matrix without a constant.  
    //covar_mxd.rows() == number of covariates (no constant)
    //covar_mxd.cols() == number of subjects
    //post: covar_with_constant_mxd.rows() == number of covariates (including
    //constant term)
    //covar_with_constant_mxd.cols() == number of cases.
    //covar_with_constant_mxd.row( 0 ) == a row of ones.
    //covar_with_constant_mxd.row( 1 ) == covar_mxd.rows( 0 ), etc.
    int make_mxd_with_constant_from_covar_mxd( const MatrixXd & covar_mxd,
                                    MatrixXd & covar_with_constant_mxd );

    //make_mxd_with_constant_and_snp_from_covar_mxd()
    //return value is error_code as int.  NO_ERROR if no error.
    //pre: covar_mxd : covariate matrix without a constant.
    //covar_mxd.rows() == number of covariates (no constant)
    //covar_mxd.cols() == number of subjects
    //snp_vxd.rows() == number of subjects (snp_vxd should be a snp)
    //post: covar_with_constant_and_snp_mxd.rows() == number of covariates (including
    //constant term)
    //covar_with_constant_and_snp_mxd.cols() == number of cases.
    //covar_with_constant_and_snp_mxd.row( 0 ) == a row of ones.
    //covar_with_constant_and_snp_mxd.row( 1 ) == snp_vxd.transpose(), etc.
    //covar_with_constant_and_snp_mxd.row( 2 ) == covar_mxd.rows( 0 ), etc.
    int make_mxd_with_constant_and_snp_from_covar_mxd( 
                    const MatrixXd & covar_mxd,
                    const VectorXd & snp_vxd,
                    MatrixXd & covar_with_constant_and_snp_mxd );   
    
    //calc_z_stat_vec()
    //pre: theta_vec.rows() == se_col_vec.rows()
    //post: z_stat_vec.rows() == theta_vec.rows() == se_col_vec.rows()
    //z_stat_vec contains z_stat values for each element of 
    //dimension of the input values.
    void calc_z_stat_vec( const VectorXd & theta_vec,
                          const VectorXd & se_col_vec,
                                VectorXd & z_stat_vec );

    //calc_p_val_vec()
    //pre: z_stat_vec contains z-stat values
    //post: p_val_vec.rows() == z_stat_vec.rows()
    //p_val_vec contains p-values (chi-squared distribution)
    void calc_p_val_vec( const VectorXd & z_stat_vec, VectorXd & p_val_vec );

//****************
    void int_to_error_string( int myint, std::string & my_error_str );
    //return values
    const int NO_ERROR  = 0;
    const int DIM_ERROR = 2;
    const int ARG_ERROR = 4;
    const int NOT_CONVERGED_ERROR = 8;
    const int FILE_ERROR = 16;
    const int INVERT_ERROR = 32;


} //namespace
#endif
