#include "lower_detect_limit.h"
#include "common.h"
using namespace common;
namespace lower_detect_limit
{

    double calc_a_value( double gamma, double sigma_sqr, double tau_sqr )
    {
        const double numerator = 1.0;
        const double denominator = ( (gamma*gamma)/sigma_sqr + (1.0/tau_sqr) );
        const double rv = numerator / denominator;
        return rv;
    }
    
    double calc_b_value( double a_value, double gamma, 
                         double sigma_sqr, double tau_sqr, double y_value,
                         const VectorXd & beta, const VectorXd & ksi,
                         const VectorXd & X_vec, const VectorXd & Z_vec )
    {
        const double first_term = (a_value / tau_sqr) * ksi.dot(Z_vec);
        const double second_term = (a_value * gamma) / sigma_sqr * 
                                   ( y_value - beta.dot(X_vec) ) ;
        const double rv = first_term + second_term;     
        return rv;
    }

   void calc_expectation( double gamma, double sigma_sqr, double tau_sqr,
                             const VectorXd & Y_vec, const VectorXd & beta,
                             const VectorXd & ksi,
                             const MatrixXd & X_mat, const MatrixXd & Z_mat,
                             const std::vector<measurement> & measure_vector,
                             std::vector< double > & expectation_vector,
                             std::vector< double > & expectation_sqr_vector )
    {
        const size_t NUM_CASES = X_mat.cols();
        expectation_vector.clear();
        expectation_sqr_vector.clear();
        const double a_val = calc_a_value( gamma, sigma_sqr, tau_sqr );
        const double sqrt_a_val = sqrt( a_val );
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const measurement measure = measure_vector[ index ];
            if (measure.type == MEASURE_TYPE_KNOWN)
            {
                expectation_vector.push_back( measure.measured_value );
                expectation_sqr_vector.push_back(
                    measure.measured_value * measure.measured_value );
            }
            else
            if (measure.type == MEASURE_TYPE_MISSING_DATA )
            {
                const VectorXd X_vec = X_mat.col( index );
                const VectorXd Z_vec = Z_mat.col( index );
                const double b_val = calc_b_value(
                    a_val, gamma, sigma_sqr, tau_sqr, Y_vec( index ),
                    beta, ksi, X_vec, Z_vec );
                expectation_vector.push_back( b_val );
                expectation_sqr_vector.push_back( b_val*b_val + a_val );
            }
            else
            if (measure.type == MEASURE_TYPE_BELOW_DETECT_LIMIT)
            {
                const VectorXd X_vec = X_mat.col( index );
                const VectorXd Z_vec = Z_mat.col( index );
                const double b_val = calc_b_value(
                    a_val, gamma, sigma_sqr, tau_sqr, Y_vec( index ),
                    beta, ksi, X_vec, Z_vec );
                const double arg_val = (measure.upper_bound - b_val) / sqrt_a_val;
                normal norm_dis; //standard normal, double
                const double pdf_val = pdf(norm_dis, arg_val);
                const double cdf_val = cdf(norm_dis, arg_val);
                const double func_ratio = pdf_val / cdf_val;
                const double expect_val = 
                    b_val - sqrt_a_val * func_ratio;
                expectation_vector.push_back( expect_val );
        
                const double first_term = expect_val * expect_val;
                const double second_term = a_val *
                    (1 - (arg_val*func_ratio) - (func_ratio*func_ratio) );
                const double expect_sqr_val = first_term + second_term;
                expectation_sqr_vector.push_back( expect_sqr_val );
             } //measure.type ==  MEASURE_TYPE_BELOW_DETECT_LIMIT
            else
            if (measure.type == MEASURE_TYPE_ABOVE_DETECT_LIMIT)
            {
                const VectorXd X_vec = X_mat.col( index );
                const VectorXd Z_vec = Z_mat.col( index );
                const double b_val = calc_b_value(
                    a_val, gamma, sigma_sqr, tau_sqr, Y_vec( index ),
                    beta, ksi, X_vec, Z_vec );
                const double arg_val = ( b_val - measure.lower_bound ) / sqrt_a_val;
                normal norm_dis; //standard normal, double
                const double pdf_val = pdf(norm_dis, arg_val);
                const double cdf_val = cdf(norm_dis, arg_val);
                const double func_ratio = pdf_val / cdf_val;
                const double expect_val =
                    b_val + sqrt_a_val * func_ratio;
                expectation_vector.push_back( expect_val );

                const double first_term = expect_val * expect_val;
                const double second_term = a_val *
                    (1 - (arg_val*func_ratio) - (func_ratio*func_ratio) );
                const double expect_sqr_val = first_term + second_term;
                expectation_sqr_vector.push_back( expect_sqr_val );
            } //measure.type ==  MEASURE_TYPE_ABOVE_DETECT_LIMIT

        } // outer for loop
    } // calc_expectation


    void calc_Q_matrix( double sigma_sqr, double tau_sqr,       
                        const MatrixXd & X_mat, const MatrixXd & Z_mat,
                        const std::vector< double > & expectation_vector,
                        const std::vector< double > & expectation_sqr_vector,
                        MatrixXd & Q_mat )
    {
        const size_t NUM_COOR_X = X_mat.rows();
        const size_t NUM_COOR_Z = Z_mat.rows();
        const size_t NUM_CASES = X_mat.cols();
        Q_mat = MatrixXd::Zero( NUM_COOR_X + NUM_COOR_Z + 3, 
                                NUM_COOR_X + NUM_COOR_Z + 3 );
        MatrixXd sum_x_sqr_mat = MatrixXd::Zero( NUM_COOR_X,
                                                 NUM_COOR_X );
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const MatrixXd current_mat = X_mat.col( index ) * 
                                         X_mat.col( index ).transpose();
            sum_x_sqr_mat += current_mat;
        }
        sum_x_sqr_mat *= (1.0 / sigma_sqr );
        Q_mat.block( 0, 0, NUM_COOR_X, NUM_COOR_X ) = sum_x_sqr_mat;
        VectorXd expect_x_vector = VectorXd::Zero( NUM_COOR_X );
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            expect_x_vector += expectation_vector[ index ] *
                                        X_mat.col( index );
        }
        expect_x_vector *= (1.0 / sigma_sqr );
        Q_mat.block( 0, NUM_COOR_X, NUM_COOR_X, 1 ) = expect_x_vector;
        const MatrixXd expect_x_vector_trans = expect_x_vector.transpose();
        Q_mat.block( NUM_COOR_X, 0, 1, NUM_COOR_X ) = expect_x_vector_trans;
        double sum_expec_sqr = 0.0;
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            sum_expec_sqr += expectation_sqr_vector[ index ];
        }
        
        Q_mat( NUM_COOR_X, NUM_COOR_X ) = sum_expec_sqr / sigma_sqr;
        Q_mat( NUM_COOR_X + 1, NUM_COOR_X + 1 ) = 
                (NUM_CASES / ( 2.0 * sigma_sqr * sigma_sqr ) );
        MatrixXd sum_z_sqr_mat = MatrixXd::Zero( NUM_COOR_Z,
                                                 NUM_COOR_Z );
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const MatrixXd current_mat = Z_mat.col( index ) *
                                         Z_mat.col( index ).transpose();
            sum_z_sqr_mat += current_mat;
        }
        sum_z_sqr_mat *= (1.0/tau_sqr);
        Q_mat.block( NUM_COOR_X + 2, NUM_COOR_X + 2, NUM_COOR_Z, NUM_COOR_Z ) =
            sum_z_sqr_mat;
        Q_mat( NUM_COOR_X + NUM_COOR_Z + 2, NUM_COOR_X + NUM_COOR_Z + 2) = 
            NUM_CASES / ( 2.0 * tau_sqr * tau_sqr );
    }


    //calc_beta_gamma_vec : for Maximization step
    int calc_beta_gamma_vec( const MatrixXd & X_mat, const VectorXd & Y_vec,
                           const std::vector< double > & expectation_vector,  
                           const std::vector< double > & expectation_sqr_vector,
                           VectorXd & beta_gamma_vec )
    {
        const size_t NUM_CASES = X_mat.cols();
        const size_t NUM_COOR = X_mat.rows();
        beta_gamma_vec = VectorXd::Zero( NUM_COOR + 1 );
        VectorXd Y_expect_vec = VectorXd::Zero( NUM_COOR + 1 );
        for ( size_t case_index = 0; case_index < NUM_CASES; ++case_index )
        {
            VectorXd current_vec = VectorXd::Zero( NUM_COOR + 1 );
            VectorXd current_x_vec = X_mat.col( case_index );
            double current_y_val = Y_vec( case_index );
            current_vec.block( 0, 0, NUM_COOR, 1 ) = current_y_val  * current_x_vec;
            current_vec( NUM_COOR ) = current_y_val * expectation_vector[ case_index ];
            Y_expect_vec += current_vec;
        }
        
        MatrixXd X_expect_mat = MatrixXd::Zero( NUM_COOR + 1 , NUM_COOR + 1 );
        for ( size_t case_index = 0; case_index < NUM_CASES; ++case_index )
        {
            MatrixXd current_mat = MatrixXd::Zero( NUM_COOR + 1 , NUM_COOR + 1 );
            const MatrixXd X_sqr_mat = 
                X_mat.col( case_index ) * X_mat.col( case_index ).transpose();
            current_mat.block( 0, 0 , NUM_COOR, NUM_COOR ) = X_sqr_mat;
            current_mat.block( NUM_COOR, 0, 1, NUM_COOR ) = 
                expectation_vector[ case_index ] * X_mat.col( case_index ).transpose();
            current_mat.block( 0, NUM_COOR, NUM_COOR, 1 ) =
                expectation_vector[ case_index ] * X_mat.col( case_index );
            current_mat( NUM_COOR, NUM_COOR ) = expectation_sqr_vector[ case_index ];
            X_expect_mat += current_mat;
        } //for loop through cases
        FullPivLU<MatrixXd> lu = X_expect_mat.fullPivLu();
        if ( ! lu.isInvertible() )
        {
            std::cout << 
                "Error: calc_beta_gamma_vec : X_expect_mat is not invertible." <<
                std::endl;
            return INVERT_ERROR;
        }
        MatrixXd inv_X_expect_mat =  X_expect_mat.inverse();
        beta_gamma_vec = inv_X_expect_mat * Y_expect_vec;
        return NO_ERROR;
    } //calc_beta_gamma_vec

    //calc_sigma_sqr : for Maximization step
    double calc_sigma_sqr( double gamma, 
                           const VectorXd & Y_vec, const VectorXd & beta,
                           const MatrixXd & X_mat, 
                           const std::vector<double> & expectation_vector,
                           const std::vector<double> & expectation_sqr_vector )
    {
        double rv = 0.0;
        const size_t NUM_CASES = X_mat.cols();
        double sum = 0.0;
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const double residual = 
                ( Y_vec( index ) - beta.dot( X_mat.col( index ) ) );
            const double first_term = residual * residual;
            const double second_term = 
                gamma * gamma * expectation_sqr_vector[ index ];                      
            const double third_term = 
                (-2.0) * residual * gamma * expectation_vector[ index ];
            sum += ( first_term + second_term + third_term );
        }
        rv = sum / NUM_CASES;
        return rv;
    }  //calc_sigma_sqr

    //calc_ksi : for Maximization step
    int calc_ksi( const MatrixXd & Z_mat, 
                  const std::vector<double> & expectation_vector,
                  VectorXd & ksi )
    {
        const size_t NUM_CASES = Z_mat.cols();
        const size_t NUM_COOR = Z_mat.rows();
        VectorXd expect_Z_vec = VectorXd::Zero( NUM_COOR );
        for ( size_t case_index = 0; case_index < NUM_CASES; 
                ++case_index )
        {
            expect_Z_vec += 
              ( expectation_vector[ case_index ]  * 
                    Z_mat.col( case_index ) );
        } //for
        
        MatrixXd Z_sqr_mat = MatrixXd::Zero( NUM_COOR, NUM_COOR );
        for ( size_t case_index = 0; case_index < NUM_CASES;
                ++case_index )
        {
            Z_sqr_mat += ( Z_mat.col( case_index ) *
                           Z_mat.col( case_index ).transpose() ); 
        } //for
        FullPivLU<MatrixXd> lu = Z_sqr_mat.fullPivLu();
        if ( ! lu.isInvertible() )
        {
            std::cout << 
                "Error: calc_ksi : Z_sqr_mat is not invertible." <<
                std::endl;
            return INVERT_ERROR;
        }
        MatrixXd inv_Z_sqr_mat = Z_sqr_mat.inverse();
        ksi = inv_Z_sqr_mat * expect_Z_vec;

        return NO_ERROR;
    } //calc_ksi


    //calc_tau_sqr : for Maximization step
    double calc_tau_sqr( const VectorXd & ksi, const MatrixXd & Z_mat,
                        const std::vector<double> & expectation_vector,
                        const std::vector<double> & expectation_sqr_vector )
    {
        double rv = 0.0;
        const size_t NUM_CASES = Z_mat.cols();
        double sum = 0.0;
        for ( size_t case_index = 0; case_index < NUM_CASES; 
                ++case_index )
        {
            const double first_term = expectation_sqr_vector[ case_index ];
            const double dot_product_val = ksi.dot(Z_mat.col( case_index ));
            const double second_term = dot_product_val * dot_product_val;
            const double third_term = 
                (-2.0)*dot_product_val*expectation_vector[ case_index ];
            sum += (first_term + second_term + third_term);
        } //for
        rv = sum / NUM_CASES;
        return rv;
    } //calc_tau_sqr

    int perform_EM( double & gamma, double & sigma_sqr,
                    double & tau_sqr,
                    VectorXd & beta, VectorXd & ksi,
                    const VectorXd & Y_vec,
                    const MatrixXd & X_mat,
                    const MatrixXd & Z_mat,
                    const std::vector<measurement> & measure_vector,
                    VectorXd & se_vec,
                    size_t MAX_ITER,
                    double CONVERGED_RATIO,
                    MatrixXd & covariance_mat,
                    VectorXd & full_estimate_vec
                   )
    {
        const size_t NUM_CASES = X_mat.cols();
        if ( ( Z_mat.cols() != NUM_CASES ) ||
             ( Y_vec.rows() != NUM_CASES ) ||
             ( measure_vector.size() != NUM_CASES ) )
        {
            std::cout << "error : perform_EM() : dimension error." << std::endl;
            std::cout << "X_mat.cols() == " << X_mat.cols() << std::endl;
            std::cout << "Z_mat.cols() == " << Z_mat.cols() << std::endl;
            std::cout << "Y_vec.rows() == " << Y_vec.rows() << std::endl;
            return DIM_ERROR;
        }

        const size_t NUM_COOR_X = X_mat.rows();
        const size_t NUM_COOR_Z = Z_mat.rows();
        static bool print_info = true;
        if ( print_info )
        {
            std::cout << "Continuous phenotype : The following settings are being used : " 
                      << std::endl;
            std::cout << "MAX_ITER == "                    
                  << MAX_ITER << std::endl;
            std::cout << "CONVERGED_RATIO == "               
                  <<  CONVERGED_RATIO << std::endl; 
            print_info = false;
        }
        //***set initial values
        gamma = 0.0; 
        beta = VectorXd::Zero( NUM_COOR_X );
        ksi = VectorXd::Zero( NUM_COOR_Z );
        std::vector<double> Y_double_std_vec;
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            Y_double_std_vec.push_back( Y_vec( index ) );
        } //for
        sigma_sqr = calc_variance( Y_double_std_vec );
        std::vector<double> measure_known_std_vec;
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            if ( measure_vector[ index ].type == MEASURE_TYPE_KNOWN )
            {
                measure_known_std_vec.push_back( 
                    measure_vector[ index ].measured_value );
            } //if
        } //for
        tau_sqr = calc_variance( measure_known_std_vec );

        //***EM loop : for each iteration first perform expectation 
        //calculations then maximization calculations  
        bool isConverged = false;    
        std::vector< double > expectation_vector;
        std::vector< double > expectation_sqr_vector;
        for ( size_t iter = 0; iter < MAX_ITER; ++iter )
        {
            const double old_gamma = gamma;
            const double old_sigma_sqr = sigma_sqr;
            const double old_tau_sqr = tau_sqr;
            const VectorXd old_beta = beta;
            const VectorXd old_ksi = ksi;
            //***Expectation step
            calc_expectation( gamma, sigma_sqr, tau_sqr,
                    Y_vec, beta, 
                    ksi,
                    X_mat, Z_mat,
                    measure_vector, 
                    expectation_vector,
                    expectation_sqr_vector );
            //***Maximization Step
            VectorXd beta_gamma_vec;
            int beta_gamma_rv = calc_beta_gamma_vec(
                    X_mat, Y_vec, 
                    expectation_vector, expectation_sqr_vector,
                    beta_gamma_vec );
            if ( beta_gamma_rv != NO_ERROR )
            {
                std::string my_error_str;
                int_to_error_string( beta_gamma_rv, my_error_str );
                std::cout << "perform_EM() : call to calc_beta_gamma returned " <<
                         my_error_str << std::endl;
                return beta_gamma_rv;
            }
            //updating beta and gamma
            gamma = beta_gamma_vec( NUM_COOR_X );
            beta = beta_gamma_vec.block( 0, 0, NUM_COOR_X, 1 ); 

            //updating sigma_sqr
            sigma_sqr = calc_sigma_sqr( 
                gamma, Y_vec, beta, 
                X_mat,
                expectation_vector, expectation_sqr_vector );
            
            //updating ksi
            int ksi_rv = calc_ksi( Z_mat,
                expectation_vector, ksi );
            if ( ksi_rv != NO_ERROR )
            {
                std::string my_error_str;
                int_to_error_string( ksi_rv, my_error_str );
                std::cout << "error : perform_EM() returned " << 
                   my_error_str << std::endl;
                return ksi_rv;
            } 
            //updating tau_sqr
            tau_sqr = calc_tau_sqr( 
                ksi, Z_mat, 
                expectation_vector, expectation_sqr_vector );
            
            //***Check for convergence
            VectorXd old_theta = VectorXd::Zero( 
                                    NUM_COOR_X + NUM_COOR_Z + 3 );
            old_theta.block( 0, 0, NUM_COOR_X, 1 ) = old_beta;
            old_theta(  NUM_COOR_X ) = old_gamma;
            old_theta( NUM_COOR_X + 1 ) = old_sigma_sqr;
            old_theta.block( NUM_COOR_X + 2, 0, NUM_COOR_Z, 1 ) = old_ksi;
            old_theta( NUM_COOR_X + NUM_COOR_Z + 2 ) = old_tau_sqr;
            
            VectorXd new_theta = VectorXd::Zero(             
                                    NUM_COOR_X + NUM_COOR_Z + 3 );
            new_theta.block( 0, 0, NUM_COOR_X, 1 ) = beta;
            new_theta(  NUM_COOR_X ) = gamma;     
            new_theta( NUM_COOR_X + 1 ) = sigma_sqr;          
            new_theta.block( NUM_COOR_X + 2, 0, NUM_COOR_Z, 1 ) = ksi;
            new_theta( NUM_COOR_X + NUM_COOR_Z + 2 ) = tau_sqr;  
            isConverged = isVectorConverged( old_theta, new_theta, CONVERGED_RATIO );
            if ( isConverged )
            {
                full_estimate_vec = new_theta;
                break;
            }

        } //for iter loop

        if ( ! isConverged )
        {
            std::cout << "perform_EM() : Finished EM : Not converged." << std::endl;
            return NOT_CONVERGED_ERROR;
        }
        
        const int se_rv = calc_se_vec( gamma, sigma_sqr, tau_sqr,
                                       Y_vec,
                                       beta, ksi,
                                       X_mat, Z_mat,
                                       measure_vector,
                                       se_vec,
                                       covariance_mat
                                     );
        
        if ( se_rv != NO_ERROR )
        {
            return se_rv;
        }
        return NO_ERROR;
    } //perform_EM


    int calc_se_vec( double gamma, double sigma_sqr, double tau_sqr,
                     const VectorXd & Y_vec,
                     const VectorXd & beta, const VectorXd & ksi,
                     const MatrixXd & X_mat, const MatrixXd & Z_mat,
                     const std::vector<measurement> & measure_vector,
                     VectorXd & se_vec,
                     MatrixXd & covariance_mat
                   )
    {
        const size_t NUM_COOR_X = X_mat.rows();
        const size_t NUM_COOR_Z = Z_mat.rows();
        const size_t NUM_CASES = X_mat.cols();
        const double a_val = calc_a_value( gamma, sigma_sqr, tau_sqr );
        std::vector< MatrixXd > mu_matrix_vec;
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            MatrixXd current_mu_matrix = MatrixXd::Zero( 3, 3 );
            if ( measure_vector[ index ].type == MEASURE_TYPE_KNOWN )
            {
                mu_matrix_vec.push_back( current_mu_matrix );
                continue;
            }
            
            const VectorXd X_vec = X_mat.col( index );
            const VectorXd Z_vec = Z_mat.col( index );
            const double b_val = calc_b_value( 
                a_val, gamma, sigma_sqr, tau_sqr, Y_vec( index ),     
                beta, ksi, X_vec, Z_vec );
            
            if ( measure_vector[ index ].type == 
                    MEASURE_TYPE_MISSING_DATA )
            {
                const double mu_1 = b_val;
                const double mu_2 = b_val * b_val + a_val;
                const double mu_3 = 3*a_val*b_val + pow(b_val, 3);
                const double mu_4 = 
                    3*a_val*a_val + 6*a_val*b_val*b_val + pow(b_val, 4);
                current_mu_matrix( 0, 0 ) = 1.0;
                current_mu_matrix( 0, 1 ) = mu_1;
                current_mu_matrix( 0, 2 ) = mu_2;
                current_mu_matrix( 1, 0 ) = mu_1;
                current_mu_matrix( 1, 1 ) = mu_2;
                current_mu_matrix( 1, 2 ) = mu_3;
                current_mu_matrix( 2, 0 ) = mu_2;
                current_mu_matrix( 2, 1 ) = mu_3;
                current_mu_matrix( 2, 2 ) = mu_4;
                mu_matrix_vec.push_back( current_mu_matrix );
            }
            else
            if (  measure_vector[ index ].type == 
                    MEASURE_TYPE_BELOW_DETECT_LIMIT )
            {
                const measurement measure = measure_vector[ index ];
                const double sqrt_a_val = sqrt( a_val );       
                const double arg_val = 
                    (measure.upper_bound - b_val) / sqrt_a_val;
                normal norm_dis; //standard normal, double        
                const double pdf_val = pdf(norm_dis, arg_val);    
                const double cdf_val = cdf(norm_dis, arg_val);
                const double func_ratio = pdf_val / cdf_val;
                double f_arr[ 5 ];
                f_arr[ 0 ] = 1.0;
                f_arr[ 1 ] = -func_ratio;
                for ( int j = 2; j < 5; ++j )
                {
                    const double second_term = (j-1)*f_arr[j-2];
                    const double first_term = 
                        - ( pow(arg_val, j-1)  * func_ratio );
                    f_arr[ j ] = first_term + second_term;
                }
                double mu_arr[5];
                //mu_arr[0] is unused
                for ( int k = 1; k < 5; ++k )
                {
                    mu_arr[ k ] = 0.0;
                    for ( int j = 0; j <= k; ++j )
                    {
                        const double k_choose_j =
                            boost::math::factorial<double>( k ) /
                            ( boost::math::factorial<double>( j ) *
                              boost::math::factorial<double>( k-j ) );
                        const double a_b_term = 
                            pow( a_val, j/2.0) * pow( b_val, k-j);
                        mu_arr[k] += (k_choose_j * a_b_term * f_arr[ j ]);           
                    } //for j loop
                } //for k loop
                const double mu_1 = mu_arr[1];
                const double mu_2 = mu_arr[2];
                const double mu_3 = mu_arr[3];
                const double mu_4 = mu_arr[4];
                current_mu_matrix( 0, 0 ) = 1.0;
                current_mu_matrix( 0, 1 ) = mu_1;
                current_mu_matrix( 0, 2 ) = mu_2;
                current_mu_matrix( 1, 0 ) = mu_1;
                current_mu_matrix( 1, 1 ) = mu_2;
                current_mu_matrix( 1, 2 ) = mu_3;
                current_mu_matrix( 2, 0 ) = mu_2;
                current_mu_matrix( 2, 1 ) = mu_3;
                current_mu_matrix( 2, 2 ) = mu_4;
                mu_matrix_vec.push_back( current_mu_matrix );
            } // MEASURE_TYPE_BELOW_DETECT_LIMIT     
            else
            if (  measure_vector[ index ].type == 
                    MEASURE_TYPE_ABOVE_DETECT_LIMIT )
            {
                const measurement measure = measure_vector[ index ];
                const double sqrt_a_val = sqrt( a_val );          
                const double arg_val = 
                    (b_val - measure.lower_bound) / sqrt_a_val;
                normal norm_dis; //standard normal, double        
                const double pdf_val = pdf(norm_dis, arg_val);    
                const double cdf_val = cdf(norm_dis, arg_val);
                const double func_ratio = pdf_val / cdf_val;
                double f_arr[ 5 ];
                f_arr[ 0 ] = 1.0;
                f_arr[ 1 ] = -func_ratio;
                for ( int j = 2; j < 5; ++j )
                {
                    const double second_term = (j-1)*f_arr[j-2];
                    const double first_term = 
                        - ( pow(arg_val, j-1)  * func_ratio );
                    f_arr[ j ] = first_term + second_term;
                }
                double mu_arr[5];
                //mu_arr[0] is unused
                for ( int k = 1; k < 5; ++k )
                {
                    mu_arr[ k ] = 0.0;
                    for ( int j = 0; j <= k; ++j )
                    {
                        const double k_choose_j =
                            boost::math::factorial<double>( k ) /
                            ( boost::math::factorial<double>( j ) *
                              boost::math::factorial<double>( k-j ) );
                        const double a_b_term = 
                            pow( a_val, j/2.0) * pow( -b_val, k-j);
                        mu_arr[k] += (k_choose_j * a_b_term * f_arr[ j ]);           
                    } //for j loop
                    if ( ( k % 2 ) == 1 )
                    {
                        mu_arr[k] = -mu_arr[k];
                    }
                } //for k loop
                const double mu_1 = mu_arr[1];
                const double mu_2 = mu_arr[2];
                const double mu_3 = mu_arr[3];
                const double mu_4 = mu_arr[4];
                current_mu_matrix( 0, 0 ) = 1.0;
                current_mu_matrix( 0, 1 ) = mu_1;
                current_mu_matrix( 0, 2 ) = mu_2;
                current_mu_matrix( 1, 0 ) = mu_1;
                current_mu_matrix( 1, 1 ) = mu_2;
                current_mu_matrix( 1, 2 ) = mu_3;
                current_mu_matrix( 2, 0 ) = mu_2;
                current_mu_matrix( 2, 1 ) = mu_3;
                current_mu_matrix( 2, 2 ) = mu_4;
                mu_matrix_vec.push_back( current_mu_matrix );
            } // MEASURE_TYPE_ABOVE_DETECT_LIMIT            
        } //for index < NUM_CASES
        
        std::vector<MatrixXd> V_mat_vec;
        calc_V_mat_vec( gamma, sigma_sqr, tau_sqr, 
                        Y_vec, beta, ksi,
                        X_mat, Z_mat, V_mat_vec );
        std::vector<VectorXd> expect_U_vector_vec;
        std::vector<MatrixXd> expect_U_sqr_mat_vec;
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const MatrixXd mu_mat = mu_matrix_vec[index];
            const VectorXd mu_vec = mu_mat.col(0);
            const MatrixXd V_mat = V_mat_vec[index];
            expect_U_vector_vec.push_back( V_mat * mu_vec );
            expect_U_sqr_mat_vec.push_back( V_mat * mu_mat * (V_mat.transpose()) );
        }
                        
        //call calc_expectation function before calculating Q: 
        //this is called (rather than using mu_matrix) since mu_matrix has
        //all-zero matrices for known measurements.
        std::vector< double > expectation_vector;
        std::vector< double > expectation_sqr_vector;
        calc_expectation( gamma, sigma_sqr, tau_sqr,
                          Y_vec, beta, ksi,
                          X_mat, Z_mat, 
                          measure_vector,
                          expectation_vector,
                          expectation_sqr_vector );
        MatrixXd Q_mat;
        calc_Q_matrix( sigma_sqr, tau_sqr,               
                       X_mat, Z_mat,
                       expectation_vector,
                       expectation_sqr_vector,
                       Q_mat );

        MatrixXd sum_expect_U_mat = 
                MatrixXd::Zero( NUM_COOR_X + NUM_COOR_Z + 3,
                                NUM_COOR_X + NUM_COOR_Z + 3 );
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            if ( measure_vector[ index ].type == MEASURE_TYPE_KNOWN )
            {
                continue; //skip known measurements
            }
            const MatrixXd current_mat = 
                expect_U_sqr_mat_vec[ index ] -
                expect_U_vector_vec[ index ] * 
                expect_U_vector_vec[ index ].transpose();
            sum_expect_U_mat += current_mat;
        }
        const MatrixXd Omega_mat = Q_mat - sum_expect_U_mat;

        FullPivLU<MatrixXd> lu = Omega_mat.fullPivLu();

        if ( ! lu.isInvertible() )
        {
            std::cout <<
                "Error: calc_se_vec : Omega_mat is not invertible." <<
                std::endl;
            std::cout <<
                "Rank of Omega_mat == " << lu.rank() << std::endl;
            std::cout <<
                "Rows of Omega_mat == " << Omega_mat.rows() << std::endl;
            std::cout <<
                "Cols of Omega_mat == " << Omega_mat.cols() << std::endl;
            return INVERT_ERROR;
        }
        
        covariance_mat = Omega_mat.inverse();
        const VectorXd variance_vec = covariance_mat.diagonal();
        se_vec = VectorXd::Zero( NUM_COOR_X + NUM_COOR_Z + 3 );
        for ( size_t index = 0; index < (NUM_COOR_X + NUM_COOR_Z + 3);
                ++index )
        {
            se_vec( index ) = sqrt( variance_vec( index ) );
        }  
        return NO_ERROR;
    } //calc_se_vec

    void calc_V_mat_vec( double gamma, double sigma_sqr, double tau_sqr,
                        const VectorXd & Y_vec, 
                        const VectorXd & beta,
                        const VectorXd & ksi,
                        const MatrixXd & X_mat,
                        const MatrixXd & Z_mat,   
                        std::vector< MatrixXd >  & V_mat_vec )
    {
        const size_t NUM_COOR_X = X_mat.rows();
        const size_t NUM_COOR_Z = Z_mat.rows();
        const size_t NUM_CASES = X_mat.cols();
        for ( size_t index = 0; index < NUM_CASES; ++index )
        {
            const double current_Y_val = Y_vec( index );
            const VectorXd current_X_vec = X_mat.col( index );
            const VectorXd current_Z_vec = Z_mat.col( index );
            const double current_residual = 
                ( current_Y_val - beta.dot( current_X_vec ) );
            MatrixXd current_V_mat = 
                MatrixXd::Zero( NUM_COOR_X + NUM_COOR_Z + 3, 3 );
            //first column = 0
            current_V_mat.block( 0, 0, NUM_COOR_X, 1 ) = 
                current_residual * 
                (current_X_vec / sigma_sqr);
            //current_V_mat( NUM_COOR_X, 0 ) = 0.0;
            current_V_mat( NUM_COOR_X + 1, 0 ) = 
                (-0.5 / sigma_sqr) + 
                0.5 * (current_residual * current_residual ) / 
                    (sigma_sqr * sigma_sqr ); 
            current_V_mat.block( NUM_COOR_X + 2, 0, NUM_COOR_Z, 1 ) =
                -(1.0/tau_sqr) * 
                 ksi.dot(current_Z_vec) * 
                    current_Z_vec;
            current_V_mat( NUM_COOR_X + NUM_COOR_Z + 2, 0 ) =
                (-0.5 / tau_sqr ) + 
                (0.5/(tau_sqr*tau_sqr)) * 
                ( ksi.dot( current_Z_vec ) * ksi.dot( current_Z_vec ) );
            //second_column = 1
            current_V_mat.block( 0, 1, NUM_COOR_X, 1 ) =
                -(gamma/sigma_sqr) * current_X_vec;
            current_V_mat( NUM_COOR_X, 1 ) = 
                (1.0/sigma_sqr) * current_residual;
            current_V_mat( NUM_COOR_X + 1, 1 ) =
                -(gamma/(sigma_sqr*sigma_sqr))*current_residual;
            current_V_mat.block( NUM_COOR_X + 2, 1, NUM_COOR_Z, 1 ) = 
                current_Z_vec / tau_sqr;
            current_V_mat( NUM_COOR_X + NUM_COOR_Z + 2, 1 ) =
                -(ksi.dot(current_Z_vec)/ (tau_sqr*tau_sqr) );
            //third_column = 2
            current_V_mat( NUM_COOR_X, 2 ) =
                -gamma / sigma_sqr;
            current_V_mat( NUM_COOR_X + 1, 2 ) =
                 (0.5*(gamma*gamma) ) / (sigma_sqr * sigma_sqr );
            current_V_mat( NUM_COOR_X + NUM_COOR_Z + 2, 2 ) =
                0.5 / (tau_sqr * tau_sqr );
    
            //add current matrix to result std::vector 
            V_mat_vec.push_back( current_V_mat );        
        }
    } //calc_V_mat_vec


} //namespace lower_detect_limit



