#ifndef LOWER_DETECT_LIMIT_H
#define LOWER_DETECT_LIMIT_H
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <eigen3/Eigen/Dense>
#include <boost/math/distributions/normal.hpp>
  using boost::math::normal; // default type is double.
#include <boost/math/special_functions/factorials.hpp>
#include <cfloat> //DBL_MIN
#include "common.h"
using namespace common;
using Eigen::MatrixXd;
using Eigen::VectorXd;

//for parameter estimation for continuous phenotype
//see perform_EM()
namespace lower_detect_limit
{

    double calc_a_value( double gamma, double sigma_sqr, double tau_sqr );

    double calc_b_value( double a_value, double gamma, 
                         double sigma_sqr, double tau_sqr, double y_value,
                         const VectorXd & beta, const VectorXd & ksi, 
                         const VectorXd & X_vec, const VectorXd & Z_vec );
    void calc_expectation( double gamma, double sigma_sqr, double tau_sqr,
                             const VectorXd & Y_vec, const VectorXd & beta,
                             const VectorXd & ksi,
                             const MatrixXd & X_mat, const MatrixXd & Z_mat,
                             const std::vector<measurement> & measure_vector,
                             std::vector< double > & expectation_vector,
                             std::vector< double > & expectation_sqr_vector ); 
    
    void calc_Q_matrix( double sigma_sqr, double tau_sqr,
                        const MatrixXd & X_mat, const MatrixXd & Z_mat,
                        const std::vector< double > & expectation_vector,
                        const std::vector< double > & expectation_sqr_vector,
                        MatrixXd & Q_mat );

    void calc_V_mat_vec( double gamma, double sigma_sqr, double tau_sqr,     
                        const VectorXd & Y_vec,
                        const VectorXd & beta,             
                        const VectorXd & ksi,
                        const MatrixXd & X_mat,
                        const MatrixXd & Z_mat,                 
                        std::vector< MatrixXd >  & V_mat_vec );

    //calc_beta_gamma_vec : for Maximization step
    int calc_beta_gamma_vec( const MatrixXd & X_mat, const VectorXd & Y_vec,
                           const std::vector< double > & expectation_vector,
                           const std::vector< double > & expectation_sqr_vector, 
                           VectorXd & beta_gamma_vec );

    //calc_sigma_sqr : for Maximization step
    double calc_sigma_sqr( double gamma, 
                           const VectorXd & Y_vec, const VectorXd & beta,
                           const MatrixXd & X_mat,
                           const std::vector<double> & expectation_vector,
                           const std::vector<double> & expectation_sqr_vector );

    //calc_ksi : for Maximization step
    int calc_ksi( const MatrixXd & Z_mat, const std::vector<double> & expectation_vector,
                  VectorXd & ksi );

    //calc_tau_sqr : for Maximization step
    double calc_tau_sqr( const VectorXd & ksi, const MatrixXd & Z_mat, 
                        const std::vector<double> & expectation_vector,
                        const std::vector<double> & expectation_sqr_vector );  

    //performs parameter estimation using an EM
    //algorithm.
    //return value: error code as int.  NO_ERROR if no error.
    //NOT_CONVERGED_ERROR if algorithm didd not converge.
    //pre: Y_vec is a continuous phenotype.
    //Y_vec.rows() == number of subjects (no missing phenotype).
    //X_mat == matrix of covariates (including constant)
    //X_mat.rows() == number of covariates (including constant)
    //The SNP may or may not be included as a row in X_mat.
    //X_mat.cols() == number of subjects (no missing phenotype)
    //Z_mat == matrix of covariates (including constant)
    //Z_mat.rows() == number of covariates (including constant)
    //The SNP will probably be included as a row in Z_mat
     //(but it is ultimately optional as far as this method is
    //concerned - Z_mat should be set before calling this method).
    //Z_mat.cols() == number of subjects (no missing phenotype)
    //measure_vector : std::vector<measurement> used
    //for omics measurement.
    //measure_vector.size() == number of subjects
    //(no missing phenotype)
    //MAX_ITER : number of iterations allowed for the EM algorithm
    //CONVERGED_RATIO : double value passed from
    //this method to isVectorConverged()
    //post: All the non-constant reference arguments
    //will contain estimates.
    int perform_EM( double & gamma, double & sigma_sqr, 
                    double & tau_sqr,
                    VectorXd & beta, VectorXd & ksi,
                    const VectorXd & Y_vec, 
                    const MatrixXd & X_mat,
                    const MatrixXd & Z_mat,
                    const std::vector<measurement> & measure_vector,
                    VectorXd & se_vec,
                    size_t MAX_ITER,
                    double CONVERGED_RATIO,
                    MatrixXd & covariance_mat,
                    VectorXd & full_estimate_vec
                    );


    int calc_se_vec( double gamma, double sigma_sqr, double tau_sqr,
                     const VectorXd & Y_vec, 
                     const VectorXd & beta, const VectorXd & ksi,
                     const MatrixXd & X_mat, const MatrixXd & Z_mat,
                     const std::vector<measurement> & measure_vector,
                     VectorXd & se_vec,
                     MatrixXd & covariance_mat
                     );

} //namespace lower_detect_limit
#endif
